#ifndef THCP_H
#define THCP_H

#include <torch/csrc/python_headers.h>
#include <TH/TH.h>
#include <THC/THC.h>
#include <TH/THHalf.h>
#include <THC/THCTensor.hpp>

#include <torch/csrc/THP.h>
#include <torch/csrc/cuda/serialization.h>
#include <torch/csrc/cuda/Module.h>
#include <torch/csrc/cuda/Storage.h>
#include <torch/csrc/cuda/Stream.h>
#include <torch/csrc/cuda/Event.h>
#include <torch/csrc/cuda/utils.h>

#endif
