#pragma once

#include <torch/data/datasets/base.h>
#include <torch/data/datasets/chunk.h>
#include <torch/data/datasets/map.h>
#include <torch/data/datasets/mnist.h>
#include <torch/data/datasets/shared.h>
#include <torch/data/datasets/stateful.h>
#include <torch/data/datasets/tensor.h>
